from __future__ import absolute_import, division, print_function

import pytest
from satyr.binpack import bf, bfd, ff, ffd, mr, weight
from satyr.proxies.messages import Cpus, Mem, Offer, TaskInfo


@pytest.fixture
def offers():
    resources = [(1.1, 2048),
                 (2.0, 512),
                 (0.8, 1024),
                 (1.6, 2048)]
    return [Offer(resources=[Cpus(cpus), Mem(mem)])
            for cpus, mem in resources]


@pytest.fixture
def tasks():
    resources = [(0.1, 128),
                 (1.0, 256),
                 (3.0, 4096),
                 (0.2, 64),
                 (1.4, 1024),
                 (0.5, 128),
                 (0.1, 128)]

    return [TaskInfo(resources=[Cpus(cpus), Mem(mem)])
            for cpus, mem in resources]


def test_weight(offers):
    assert round(sum(weight(offers, cpus=1)), 1) == 1.0
    assert round(sum(weight(offers, cpus=1, mem=1)), 1) == 2.0
    assert round(sum(weight(offers, cpus=4, mem=1.5)), 1) == 5.5


def test_ff(tasks, offers):
    bins, skip = ff(tasks, offers)
    assert skip == [tasks[2]]


def test_ffd(tasks, offers):
    bins, skip = ffd(tasks, offers, cpus=1, mem=1)
    assert skip == [tasks[2]]


def test_mr(tasks, offers):
    bins, skip = mr(tasks, offers, cpus=1, mem=1)
    assert skip == [tasks[2], tasks[4]]


def test_bf(tasks, offers):
    bins, skip = bf(tasks, offers, cpus=1, mem=1)
    assert skip == [tasks[2]]


def test_bfd(tasks, offers):
    bins, skip = bfd(tasks, offers, cpus=1, mem=1)
    assert skip == [tasks[2]]



from __future__ import absolute_import, division, print_function

import operator


def weight(items, **kwargs):
    if not len(kwargs):
        raise ValueError('Missing attribute for weighting items!')
    scaled = []
    for attr, weight in kwargs.items():
        values = [float(getattr(item, attr)) for item in items]
        try:
            s = sum(values)
            scaled.append([weight * (v / s) for v in values])
        except ZeroDivisionError:
            # s equals to zero, attr wont contribute
            scaled.append([0] * len(items))

    return map(sum, zip(*scaled))


def ff(items, targets):
    """First-Fit

    This is perhaps the simplest packing heuristic;
    it simply packs items in the next available bin.

    Complexity O(n^2)
    """
    bins = [(target, []) for target in targets]
    skip = []

    for item in items:
        for target, content in bins:
            if item <= (target - sum(content)):
                content.append(item)
                break
        else:
            skip.append(item)
    return bins, skip


def ffd(items, targets, **kwargs):
    """First-Fit Decreasing

    This is perhaps the simplest packing heuristic;
    it simply packs items in the next available bin.

    This algorithm differs only from Next-Fit Decreasing
    in having a 'sort'; that is, the items are pre-sorted
    (largest to smallest).

    Complexity O(n^2)
    """
    sizes = zip(items, weight(items, **kwargs))
    sizes = sorted(sizes, key=operator.itemgetter(1), reverse=True)
    items = map(operator.itemgetter(0), sizes)
    return ff(items, targets)


def mr(items, targets, **kwargs):
    """Max-Rest

    Complexity O(n^2)
    """
    bins = [(target, []) for target in targets]
    skip = []

    for item in items:
        capacities = [target - sum(content) for target, content in bins]
        weighted = weight(capacities, **kwargs)

        (target, content), capacity, _ = max(zip(bins, capacities, weighted),
                                             key=operator.itemgetter(2))
        if item <= capacity:
            content.append(item)
        else:
            skip.append(item)
    return bins, skip


def mrpq(items, targets):
    """Max-Rest Priority Queue

    Complexity O(n*log(n))
    """
    raise NotImplementedError()


def bf(items, targets, **kwargs):
    """Best-Fit

    Complexity O(n^2)
    """
    bins = [(target, []) for target in targets]
    skip = []

    for item in items:
        containers = []
        capacities = []
        for target, content in bins:
            capacity = target - sum(content)
            if item <= capacity:
                containers.append(content)
                capacities.append(capacity - item)

        if len(capacities):
            weighted = zip(containers, weight(capacities, **kwargs))
            content, _ = min(weighted, key=operator.itemgetter(1))
            content.append(item)
        else:
            skip.append(item)
    return bins, skip


def bfd(items, targets, **kwargs):
    """Best-Fit Decreasing

    Complexity O(n^2)
    """
    sizes = zip(items, weight(items, **kwargs))
    sizes = sorted(sizes, key=operator.itemgetter(1), reverse=True)
    items = map(operator.itemgetter(0), sizes)
    return bf(items, targets, **kwargs)


def bfh(items, targets):
    """Best-Fit-Heap

    Slightly Improved Complexity
    """
    raise NotImplementedError()



from __future__ import absolute_import, division, print_function

import pytest
from sample_pb2 import MessageOfTypes
from satyr.protobuf import dict_to_protobuf, protobuf_to_dict


@pytest.fixture
def m():
    m = MessageOfTypes()
    m.dubl = 1.7e+308
    m.flot = 3.4e+038
    m.i32 = 2 ** 31 - 1  # 2147483647 #
    m.i64 = 2 ** 63 - 1  # 0x7FFFFFFFFFFFFFFF
    m.ui32 = 2 ** 32 - 1
    m.ui64 = 2 ** 64 - 1
    m.si32 = -1 * m.i32
    m.si64 = -1 * m.i64
    m.f32 = m.i32
    m.f64 = m.i64
    m.sf32 = m.si32
    m.sf64 = m.si64
    m.bol = True
    m.strng = "string"
    m.byts = b'\n\x14\x1e'
    assert len(m.byts) == 3, len(m.byts)
    m.nested.req = "req"
    m.enm = MessageOfTypes.C  # @UndefinedVariable
    m.enmRepeated.extend([MessageOfTypes.A, MessageOfTypes.C])
    m.range.extend(range(10))
    return m


def compare(m, d, exclude=None):
    i = 0
    exclude = ['byts', 'nested', 'enm', 'enmRepeated'] + (exclude or [])
    for i, field in enumerate(MessageOfTypes.DESCRIPTOR.fields):  # @UndefinedVariable
        if field.name not in exclude:
            assert field.name in d, field.name
            assert d[field.name] == getattr(
                m, field.name), (field.name, d[field.name])
    assert i > 0
    assert m.byts == str(d['byts'])
    assert d['nested'] == {'req': m.nested.req}


def test_basics(m):
    d = protobuf_to_dict(m)
    compare(m, d, ['nestedRepeated'])

    m2 = dict_to_protobuf(d, MessageOfTypes)
    assert m == m2


def test_use_enum_labels(m):
    d = protobuf_to_dict(m)
    compare(m, d, ['nestedRepeated'])
    assert d['enm'] == 'C'
    assert d['enmRepeated'] == ['A', 'C']

    m2 = dict_to_protobuf(d, MessageOfTypes)
    assert m == m2

    d['enm'] = 'MEOW'
    with pytest.raises(KeyError):
        dict_to_protobuf(d, MessageOfTypes)

    d['enm'] = 'A'
    d['enmRepeated'] = ['B']
    dict_to_protobuf(d, MessageOfTypes)

    d['enmRepeated'] = ['CAT']
    with pytest.raises(KeyError):
        dict_to_protobuf(d, MessageOfTypes)


def test_repeated_enum(m):
    d = protobuf_to_dict(m)
    compare(m, d, ['nestedRepeated'])
    assert d['enmRepeated'] == ['A', 'C']

    m2 = dict_to_protobuf(d, MessageOfTypes)
    assert m == m2

    d['enmRepeated'] = ['MEOW']
    with pytest.raises(KeyError):
        dict_to_protobuf(d, MessageOfTypes)


def test_nested_repeated(m):
    m.nestedRepeated.extend(
        [MessageOfTypes.NestedType(req=str(i)) for i in range(10)])

    d = protobuf_to_dict(m)
    compare(m, d, exclude=['nestedRepeated'])
    assert d['nestedRepeated'] == [{'req': str(i)} for i in range(10)]

    m2 = dict_to_protobuf(d, MessageOfTypes)
    assert m == m2


def test_reverse(m):
    m2 = dict_to_protobuf(protobuf_to_dict(m), MessageOfTypes)
    assert m == m2
    m2.dubl = 0
    assert m2 != m


def test_incomplete(m):
    d = protobuf_to_dict(m)
    d.pop('dubl')
    m2 = dict_to_protobuf(d, MessageOfTypes)
    assert m2.dubl == 0
    assert m != m2


def test_non_strict(m):
    d = protobuf_to_dict(m)
    d['non_existing_field'] = 'data'
    d['temporary_field'] = 'helping_state'

    with pytest.raises(KeyError):
        dict_to_protobuf(d, MessageOfTypes)

    m2 = dict_to_protobuf(d, MessageOfTypes, strict=False)
    with pytest.raises(AttributeError):
        m2.temporary_field


def test_pass_instance(m):
    d = protobuf_to_dict(m)
    d['dubl'] = 1
    m2 = dict_to_protobuf(d, m)
    assert m is m2
    assert m.dubl == 1


def test_container_mapping(m):
    class mapping(dict):
        pass

    containers = [(MessageOfTypes.NestedType(), mapping),
                  (MessageOfTypes(), dict)]

    m.nestedRepeated.extend([MessageOfTypes.NestedType(req='1')])
    d = protobuf_to_dict(m, containers=containers)
    m = dict_to_protobuf(d, containers=containers)

    assert isinstance(d, dict)
    assert isinstance(d['nested'], mapping)
    assert isinstance(m, MessageOfTypes)
    assert isinstance(m.nested, MessageOfTypes.NestedType)


def test_conditional_container_mapping(m):
    class truedict(dict):
        pass

    class falsedict(dict):
        pass

    containers = [(MessageOfTypes(bol=True), truedict),
                  (MessageOfTypes(bol=False), falsedict),
                  (MessageOfTypes.NestedType(), dict)]

    m.bol = True
    d = protobuf_to_dict(m, containers=containers)
    p = dict_to_protobuf(d, containers=containers)

    assert isinstance(d, truedict)
    assert isinstance(p, MessageOfTypes)

    m.bol = False
    d = protobuf_to_dict(m, containers=containers)
    p = dict_to_protobuf(d, containers=containers)

    assert isinstance(d, falsedict)
    assert isinstance(p, MessageOfTypes)



from __future__ import absolute_import, division, print_function

import signal
from contextlib import contextmanager


class TimeoutError(Exception):
    pass


@contextmanager
def timeout(seconds):
    def signal_handler(signum, frame):
        raise TimeoutError("Timed out!")

    if seconds > 0:
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(seconds)
        try:
            yield
        finally:
            signal.alarm(0)
    else:  # infinite timeout
        yield



from __future__ import absolute_import, division, print_function

import pytest
from satyr.messages import PythonTask, PythonTaskStatus
from satyr.proxies.messages import (Cpus, Disk, Mem, Offer, OfferID, SlaveID,
                                    TaskID)
from satyr.scheduler import QueueScheduler, Running


@pytest.fixture
def python_task():
    task = PythonTask(id=TaskID(value='test-task-id'),
                      fn=sum, args=[range(5)],
                      resources=[Cpus(0.1), Mem(128), Disk(0)])
    return task


@pytest.fixture
def offers():
    o1 = Offer(id=OfferID(value='first-offer'),
               slave_id=SlaveID(value='test-slave'),
               resources=[Cpus(2), Mem(256), Disk(1024)])
    o2 = Offer(id=OfferID(value='second-offer'),
               slave_id=SlaveID(value='test-slave'),
               resources=[Cpus(1), Mem(1024), Disk(2048)])
    return [o1, o2]


def test_launch_decline(mocker, python_task, offers):
    driver = mocker.Mock()
    sched = QueueScheduler(name='test-scheduler')

    sched.submit(python_task)
    sched.on_offers(driver, offers)

    calls = driver.launch.call_args_list

    args, kwargs = calls[0]
    assert isinstance(args[0], OfferID)
    assert args[0].value == 'first-offer'
    assert isinstance(args[1][0], PythonTask)
    assert args[1][0].task_id.value == 'test-task-id'

    args, kwargs = calls[1]
    assert isinstance(args[0], OfferID)
    assert args[0].value == 'second-offer'
    assert args[1] == []  # declines via launch empty task list


def test_task_callbacks(mocker, python_task, offers):
    driver = mocker.Mock()
    sched = QueueScheduler(name='test-scheduler')
    mocker.spy(python_task, 'on_update')
    mocker.spy(python_task, 'on_success')

    sched.submit(python_task)
    sched.on_offers(driver, offers)

    sched.on_update(driver, python_task.status('TASK_RUNNING'))
    sched.on_update(driver, python_task.status('TASK_FINISHED',
                                               data=python_task()))

    update_calls = python_task.on_update.call_args_list
    success_calls = python_task.on_success.call_args_list

    args, kwargs = update_calls[0]
    assert isinstance(args[0], PythonTaskStatus)
    assert args[0].state == 'TASK_RUNNING'

    args, kwargs = update_calls[1]
    assert isinstance(args[0], PythonTaskStatus)
    assert args[0].state == 'TASK_FINISHED'

    args, kwargs = success_calls[0]
    assert isinstance(args[0], PythonTaskStatus)
    assert args[0].state == 'TASK_FINISHED'
    assert args[0].data == 10


def test_task_result(mocker, python_task, offers):
    driver = mocker.Mock()
    sched = QueueScheduler(name='test-scheduler')

    result = sched.submit(python_task)
    sched.on_offers(driver, offers)
    sched.on_update(driver, python_task.status('TASK_RUNNING'))
    sched.on_update(driver, python_task.status('TASK_FINISHED',
                                               data=python_task()))

    assert result.get() == 10


# integration test
def test_runner_context_manager():
    sched = QueueScheduler(name='test-scheduler')
    with Running(sched, name='test-scheduler'):
        sched.wait()

    assert sched



from __future__ import absolute_import, division, print_function

import logging
import sys

from mesos.interface import Executor

from .messages import decode, encode


class ExecutorProxy(Executor):
    """Base class for Mesos executors.

    Users' executors should extend this class to get default implementations of
    methods they don't override.
    """

    def __init__(self, executor):
        self.executor = executor

    def registered(self, driver, executorInfo, frameworkInfo, slaveInfo):
        logging.info('Registered with slave', extra=dict())
        return self.executor.on_registered(ExecutorDriverProxy(driver),
                                           decode(executorInfo),
                                           decode(frameworkInfo),
                                           decode(slaveInfo))

    def reregistered(self, driver, slaveInfo):
        logging.info('Re-registered with slave', extra=dict())
        return self.executor.on_reregistered(ExecutorDriverProxy(driver),
                                             decode(slaveInfo))

    def disconnected(self, driver):
        logging.info('Disconnected from slave')
        return self.executor.on_disconnected(ExecutorDriverProxy(driver))

    def launchTask(self, driver, taskInfo):
        logging.info('Launches task', extra=dict())
        return self.executor.on_launch(ExecutorDriverProxy(driver),
                                       decode(taskInfo))

    def killTask(self, driver, taskId):
        logging.info('Kills task', extra=dict())
        return self.executor.on_kill(ExecutorDriverProxy(driver),
                                     decode(taskId))

    def frameworkMessage(self, driver, message):
        logging.info('Recived framework message', extra=dict())
        return self.executor.on_message(ExecutorDriverProxy(driver),
                                        message)

    def shutdown(self, driver):
        logging.info('Executor shutdown')
        return self.executor.on_shutdown(ExecutorDriverProxy(driver))

    def error(self, driver, message):
        print("Error from Mesos: %s" % message, file=sys.stderr)
        return self.executor.on_error(ExecutorDriverProxy(driver),
                                      message)


class ExecutorDriverProxy(object):

    def __init__(self, driver):
        self.driver = driver

    def start(self):
        """Starts the executor driver.

        This needs to be called before any other driver calls are made.
        """
        logging.info('Driver started')
        return self.driver.start()

    def stop(self):
        """Stops the executor driver."""
        logging.info('Driver stopped')
        return self.driver.stop()

    def abort(self):
        """Aborts the driver so that no more callbacks can be made to the
           executor.

        The semantics of abort and stop have deliberately been separated so that
        code can detect an aborted driver (i.e., via the return status of
        ExecutorDriver.join), and instantiate and start another driver if
        desired (from within the same process, although this functionality is
        currently not supported for executors).
        """
        logging.info('Driver aborted')
        return self.driver.abort()

    def join(self):
        """Waits for the driver to be stopped or aborted, possibly blocking the
           current thread indefinitely.

        The return status of this function can be used to determine if the
        driver was aborted (see mesos.proto for a description of Status).
        """
        logging.info('Joined to driver')
        return self.driver.join()

    def run(self):
        """Starts and immediately joins (i.e., blocks on) the driver."""
        logging.info('Driver run')
        return self.driver.run()

    def update(self, status):
        """Sends a status update to the framework scheduler.

        Retrying as necessary until an acknowledgement has been received or the
        executor is terminated (in which case, a TASK_LOST status update will be
        sent).
        See Scheduler.statusUpdate for more information about status update
        acknowledgements.
        """
        logging.info('Executor sends status update')
        return self.driver.sendStatusUpdate(encode(status))

    def message(self, data):
        """Sends a message to the framework scheduler.

        These messages are best effort; do not expect a framework message to be
        retransmitted in any reliable fashion.
        """
        logging.info('Driver sends framework message')
        return self.driver.sendFrameworkMessage(data)



from __future__ import absolute_import, division, print_function

# import base64
from copy import copy
from functools import partial

import six
from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.message import Message

__all__ = ["protobuf_to_dict", "TYPE_CALLABLE_MAP", "dict_to_protobuf",
           "REVERSE_TYPE_CALLABLE_MAP"]


TYPE_CALLABLE_MAP = {
    FieldDescriptor.TYPE_DOUBLE: float,
    FieldDescriptor.TYPE_FLOAT: float,
    FieldDescriptor.TYPE_INT32: int,
    FieldDescriptor.TYPE_INT64: int if six.PY3 else six.integer_types[1],
    FieldDescriptor.TYPE_UINT32: int,
    FieldDescriptor.TYPE_UINT64: int if six.PY3 else six.integer_types[1],
    FieldDescriptor.TYPE_SINT32: int,
    FieldDescriptor.TYPE_SINT64: int if six.PY3 else six.integer_types[1],
    FieldDescriptor.TYPE_FIXED32: int,
    FieldDescriptor.TYPE_FIXED64: int if six.PY3 else six.integer_types[1],
    FieldDescriptor.TYPE_SFIXED32: int,
    FieldDescriptor.TYPE_SFIXED64: int if six.PY3 else six.integer_types[1],
    FieldDescriptor.TYPE_BOOL: bool,
    FieldDescriptor.TYPE_STRING: six.text_type,
    FieldDescriptor.TYPE_BYTES: str,  # base64.b64encode,
    FieldDescriptor.TYPE_ENUM: int
}

REVERSE_TYPE_CALLABLE_MAP = {
    # FieldDescriptor.TYPE_BYTES: base64.b64decode
}

CONTAINER_MAP = []


def enum_to_label(field, value):
    return field.enum_type.values_by_number[int(value)].name


def label_to_enum(field, value):
    enum_dict = field.enum_type.values_by_name
    return enum_dict[value].number


def message_to_container(message, containers):
    for msg, cnt in containers:
        if isinstance(msg, type):  # class definition used
            if isinstance(message, msg):
                return cnt()
        elif isinstance(message, msg.__class__):  # object definition used
            if all([getattr(msg, field.name) == getattr(message, field.name)
                    for field, value in msg.ListFields()]):
                return cnt()
    return dict()  # fallback to plain dictionary


def container_to_message(container, containers):
    for msg, cnt in containers:
        if isinstance(container, cnt):
            if isinstance(msg, type):
                return msg()
            else:
                return copy(msg)


def protobuf_to_dict(pb, containers=CONTAINER_MAP, converters=TYPE_CALLABLE_MAP):
    result = message_to_container(pb, containers)

    # for field, value in pb.ListFields():  # only non-empty fields
    for field in pb.DESCRIPTOR.fields:  # empty fields too
        value = getattr(pb, field.name)
        if (field.message_type and field.message_type.has_options and
                field.message_type.GetOptions().map_entry):
            converter = dict
        elif field.type == FieldDescriptor.TYPE_MESSAGE:
            # recursively encode protobuf sub-message
            converter = partial(protobuf_to_dict, containers=containers,
                                converters=converters)
        elif field.type == FieldDescriptor.TYPE_ENUM:
            converter = partial(enum_to_label, field)
        else:
            converter = converters[field.type]

        if field.label == FieldDescriptor.LABEL_REPEATED:
            converter = partial(map, converter)

        result[field.name] = converter(value)

    return result


def dict_to_protobuf(dct, pb=None, containers=CONTAINER_MAP,
                     converters=REVERSE_TYPE_CALLABLE_MAP, strict=True):
    default = container_to_message(dct, containers)
    if pb:
        if default:
            pb.MergeFrom(default)
    else:
        pb = default
    pb = pb if isinstance(pb, Message) else pb()

    for k, v in dct.items():
        try:
            # TODO silently skip undifened fields
            field = pb.DESCRIPTOR.fields_by_name[k]
        except:
            if not strict:
                continue
            else:
                raise
        pb_value = getattr(pb, k, None)
        if field.label == FieldDescriptor.LABEL_REPEATED:
            for item in v:
                if field.type == FieldDescriptor.TYPE_MESSAGE:
                    dict_to_protobuf(item, pb_value.add(),
                                     containers, converters)
                elif field.type == FieldDescriptor.TYPE_ENUM:
                    pb_value.append(label_to_enum(field, item))
                else:
                    pb_value.append(item)
        elif field.type == FieldDescriptor.TYPE_MESSAGE:
            dict_to_protobuf(v, pb_value, containers, converters)
        else:
            if field.type in converters:
                v = converters[field.type](v)
            elif field.type == FieldDescriptor.TYPE_ENUM:
                v = label_to_enum(field, v)

            setattr(pb, field.name, v)

    return pb

encode = dict_to_protobuf
decode = protobuf_to_dict



from __future__ import absolute_import, division, print_function

import operator
from functools import partial
from uuid import uuid4

from google.protobuf.message import Message
from mesos.interface import mesos_pb2

from .. import protobuf


class Map(dict):

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self[k] = v

    @classmethod
    def cast(cls, v):
        if isinstance(v, Map):
            return v
        elif isinstance(v, dict):
            return Map(**v)
        elif hasattr(v, '__iter__'):
            return map(cls.cast, v)
        else:
            return v

    def __setitem__(self, k, v):
        # accidental __missing__ call will create a new node
        super(Map, self).__setitem__(k, self.cast(v))

    def __setattr__(self, k, v):
        prop = getattr(self.__class__, k, None)
        if isinstance(prop, property):  # property binding
            prop.fset(self, v)
        elif callable(v):  # method binding
            self.__dict__[k] = v
        else:
            self[k] = v

    def __getattr__(self, k):
        return self[k]

    # def __delattr__(self, k):
    #    del self[k]

    # def __missing__(self, k):
    #    # TODO: consider not using this, silents errors
    #    self[k] = Map()
    #    return self[k]

    def __hash__(self):
        return hash(tuple(self.items()))


class RegisterProxies(type):

    def __init__(cls, name, bases, nmspc):
        super(RegisterProxies, cls).__init__(name, bases, nmspc)
        if not hasattr(cls, 'registry'):
            cls.registry = []
        cls.registry.insert(0, (cls.proto, cls))
        # cls.registry -= set(bases) # Remove base classes

    # Metamethods, called on class objects:
    def __iter__(cls):
        return iter(cls.registry)


class MessageProxy(Map):
    __metaclass__ = RegisterProxies
    proto = Message


class Scalar(MessageProxy):
    proto = mesos_pb2.Value.Scalar


class Resource(MessageProxy):
    proto = mesos_pb2.Resource


# TODO: RangeResource e.g. ports
class ScalarResource(Resource):
    # supports comparison and basic arithmetics with scalars
    proto = mesos_pb2.Resource(type=mesos_pb2.Value.SCALAR)

    def __init__(self, value=None, **kwargs):
        super(Resource, self).__init__(**kwargs)
        if value is not None:
            self.scalar = Scalar(value=value)

    def __cmp__(self, other):
        first, second = float(self), float(other)
        if first < second:
            return -1
        elif first > second:
            return 1
        else:
            return 0

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.scalar.value)

    def __float__(self):
        return float(self.scalar.value)

    @classmethod
    def _op(cls, op, first, second):
        value = op(float(first), float(second))
        return cls(value=value)

    def __add__(self, other):
        return self._op(operator.add, self, other)

    def __radd__(self, other):
        return self._op(operator.add, other, self)

    def __sub__(self, other):
        return self._op(operator.sub, self, other)

    def __rsub__(self, other):
        return self._op(operator.sub, other, self)

    def __mul__(self, other):
        return self._op(operator.mul, self, other)

    def __rmul__(self, other):
        return self._op(operator.mul, other, self)

    def __truediv__(self, other):
        return self._op(operator.truediv, self, other)

    def __rtruediv__(self, other):
        return self._op(operator.truediv, other, self)

    def __iadd__(self, other):
        self.scalar.value = float(self._op(operator.add, self, other))
        return self

    def __isub__(self, other):
        self.scalar.value = float(self._op(operator.sub, self, other))
        return self


class Cpus(ScalarResource):
    proto = mesos_pb2.Resource(name='cpus', type=mesos_pb2.Value.SCALAR)


class Mem(ScalarResource):
    proto = mesos_pb2.Resource(name='mem', type=mesos_pb2.Value.SCALAR)


class Disk(ScalarResource):
    proto = mesos_pb2.Resource(name='disk', type=mesos_pb2.Value.SCALAR)


class ResourcesMixin(object):

    @classmethod
    def _cast_zero(cls, other=0):
        if other == 0:
            return cls(resources=[Cpus(0), Mem(0), Disk(0)])
        else:
            return other

    @property
    def cpus(self):
        for res in self.resources:
            if isinstance(res, Cpus):
                return res
        return Cpus(0.0)

    @property
    def mem(self):
        for res in self.resources:
            if isinstance(res, Mem):
                return res
        return Mem(0.0)

    @property
    def disk(self):
        for res in self.resources:
            if isinstance(res, Disk):
                return res
        return Disk(0.0)

    # @property
    # def ports(self):
    #     for res in self.resources:
    #         if isinstance(res, Ports):
    #             return [(rng.begin, rng.end) for rng in res.ranges.range]

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__,
                                 ', '.join(map(str, self.resources)))

    def __cmp__(self, other):
        other = self._cast_zero(other)

        if all([self.cpus < other.cpus,
                self.mem < other.mem,
                self.disk < other.disk]):
            # all resources are smaller the task will fit into offer
            return -1
        elif any([self.cpus > other.cpus,
                  self.mem > other.mem,
                  self.disk > other.disk]):
            # any resources is bigger task won't fit into offer
            return 1
        else:
            return 0

    def __radd__(self, other):  # to support sum()
        other = self._cast_zero(other)
        return self + other

    def __add__(self, other):
        other = self._cast_zero(other)
        # ports = list(set(self.ports) | set(other.ports))
        cpus = self.cpus + other.cpus
        mem = self.mem + other.mem
        disk = self.disk + other.disk
        mixin = self.__class__()
        mixin.resources = [cpus, disk, mem]
        return mixin

    def __sub__(self, other):
        other = self._cast_zero(other)
        # ports = list(set(self.ports) | set(other.ports))
        cpus = self.cpus - other.cpus
        mem = self.mem - other.mem
        disk = self.disk - other.disk
        mixin = self.__class__()
        mixin.resources = [cpus, disk, mem]
        return mixin

    def __iadd__(self, other):
        other = self._cast_zero(other)
        added = self + other
        self.resources = added.resources
        return self

    def __isub__(self, other):
        other = self._cast_zero(other)
        subbed = self - other
        self.resources = subbed.resources
        return self


class FrameworkID(MessageProxy):
    proto = mesos_pb2.FrameworkID


class SlaveID(MessageProxy):
    proto = mesos_pb2.SlaveID


class ExecutorID(MessageProxy):
    proto = mesos_pb2.ExecutorID


class OfferID(MessageProxy):
    proto = mesos_pb2.OfferID


class TaskID(MessageProxy):
    proto = mesos_pb2.TaskID


class FrameworkInfo(MessageProxy):
    proto = mesos_pb2.FrameworkInfo


class ExecutorInfo(MessageProxy):
    proto = mesos_pb2.ExecutorInfo


class MasterInfo(MessageProxy):
    proto = mesos_pb2.MasterInfo


class SlaveInfo(MessageProxy):
    proto = mesos_pb2.SlaveInfo


class Filters(MessageProxy):
    proto = mesos_pb2.Filters


class TaskStatus(MessageProxy):
    proto = mesos_pb2.TaskStatus

    def is_staging(self):
        return self.state == 'TASK_STAGING'

    def is_starting(self):
        return self.state == 'TASK_STARTING'

    def is_running(self):
        return self.state == 'TASK_RUNNING'

    def has_succeeded(self):
        return self.state == 'TASK_FINISHED'

    def has_failed(self):
        return self.state in ['TASK_FAILED', 'TASK_LOST', 'TASK_KILLED',
                              'TASK_ERROR']

    def has_terminated(self):
        return self.has_succeeded() or self.has_failed()


class Offer(ResourcesMixin, MessageProxy):  # important order!
    proto = mesos_pb2.Offer


class TaskInfo(ResourcesMixin, MessageProxy):
    proto = mesos_pb2.TaskInfo

    def __init__(self, id=None, **kwargs):
        super(TaskInfo, self).__init__(**kwargs)
        self.id = id or TaskID(value=str(uuid4()))

    @property
    def id(self):  # more consistent naming
        return self.task_id

    @id.setter
    def id(self, value):
        self.task_id = value

    def status(self, state, **kwargs):  # used on executor side
        return TaskStatus(task_id=self.task_id, state=state, **kwargs)


class CommandInfo(MessageProxy):
    proto = mesos_pb2.CommandInfo


class ContainerInfo(MessageProxy):
    proto = mesos_pb2.ContainerInfo


class DockerInfo(MessageProxy):
    proto = mesos_pb2.ContainerInfo.DockerInfo


class Request(MessageProxy):
    proto = mesos_pb2.Request


class Operation(MessageProxy):
    proto = mesos_pb2.Offer.Operation


decode = partial(protobuf.decode, containers=MessageProxy.registry)
encode = partial(protobuf.encode, containers=MessageProxy.registry,
                 strict=False)



from __future__ import absolute_import, division, print_function

from satyr.executor import OneOffExecutor, Running
from satyr.messages import PythonTask, PythonTaskStatus


class FakeThread(object):

    def __init__(self, target):
        self.target = target

    def start(self):
        return self.target()


def test_finished_status_updates(mocker):
    mocker.patch('threading.Thread', side_effect=FakeThread)

    driver = mocker.Mock()
    task = PythonTask(fn=sum, args=[range(5)])

    executor = OneOffExecutor()
    executor.on_launch(driver, task)

    calls = driver.update.call_args_list

    args, kwargs = calls[0]
    status = args[0]
    assert isinstance(status, PythonTaskStatus)
    assert status.state == 'TASK_RUNNING'
    assert status.data is None

    args, kwargs = calls[1]
    status = args[0]
    assert isinstance(status, PythonTaskStatus)
    assert status.state == 'TASK_FINISHED'
    assert status.data == 10


def test_failed_status_updates(mocker):
    mocker.patch('threading.Thread', FakeThread)

    def failing_function(*args):
        raise Exception("Booom!")

    driver = mocker.Mock()
    task = PythonTask(fn=failing_function, args=['arbitrary', 'args'])

    executor = OneOffExecutor()
    executor.on_launch(driver, task)

    calls = driver.update.call_args_list

    args, kwargs = calls[0]
    status = args[0]
    assert isinstance(status, PythonTaskStatus)
    assert status.state == 'TASK_RUNNING'
    assert status.data is None

    args, kwargs = calls[1]
    status = args[0]
    assert isinstance(status, PythonTaskStatus)
    assert status.state == 'TASK_FAILED'
    assert isinstance(status.data, Exception)
    assert status.message == 'Booom!'


# integration test
# def test_runner_context_manager():
#     executor = OneOffExecutor()
#     with Running(executor):
#         pass

#     assert executor



from kazoo.client import KazooClient
from kazoo.recipe.queue import LockingQueue as KazooLockingQueue
from kazoo.recipe.queue import Queue as KazooQueue


class SerializeMixin(object):

    def __getstate__(self):
        hosts = ["{}:{}".format(h, p) for h, p in self.client.hosts]
        client = ",".join(hosts)

        result = self.__dict__.copy()
        result['client'] = client

        return result

    def __setstate__(self, state):
        hosts = state.pop('client')
        client = KazooClient(hosts)
        client.start()

        self.__dict__ = state
        self.client = client


class Queue(KazooQueue, SerializeMixin):
    pass


class LockingQueue(KazooLockingQueue, SerializeMixin):
    pass



from __future__ import absolute_import, division, print_function

import atexit
import logging
import signal
import sys
import threading

from mesos.interface import mesos_pb2
from mesos.native import MesosExecutorDriver

from .interface import Executor
from .proxies import ExecutorProxy


class Running(object):

    def __init__(self, executor):
        executor = ExecutorProxy(executor)
        self.driver = MesosExecutorDriver(executor)

        def shutdown(signal, frame):
            self.driver.stop()

        signal.signal(signal.SIGINT, shutdown)
        signal.signal(signal.SIGTERM, shutdown)
        atexit.register(self.driver.stop)

    def run(self):
        return self.driver.run()

    def start(self):
        status = self.driver.start()
        assert status == mesos_pb2.DRIVER_RUNNING
        return status

    def stop(self):
        logging.info("Stopping Mesos driver")
        self.driver.stop()
        logging.info("Joining Mesos driver")
        result = self.driver.join()
        logging.info("Joined Mesos driver")
        if result != mesos_pb2.DRIVER_STOPPED:
            raise RuntimeError("Mesos driver failed with %i", result)

    def join(self):
        self.driver.join()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()


class OneOffExecutor(Executor):

    def on_launch(self, driver, task):
        def run_task():
            driver.update(task.status('TASK_RUNNING'))
            logging.info('Sent TASK_RUNNING status update')

            try:
                logging.info('Executing task...')
                result = task()
            except Exception as e:
                logging.exception('Task errored')
                driver.update(task.status('TASK_FAILED',
                                          data=e,
                                          message=e.message))
                logging.info('Sent TASK_RUNNING status update')
            else:
                driver.update(task.status('TASK_FINISHED', data=result))
                logging.info('Sent TASK_FINISHED status update')

        thread = threading.Thread(target=run_task)
        thread.start()

    def on_kill(self, driver, task_id):
        driver.stop()


if __name__ == '__main__':
    status = Running(OneOffExecutor()).run()
    code = 0 if status == mesos_pb2.DRIVER_STOPPED else 1
    sys.exit(code)



from __future__ import absolute_import, division, print_function

import time
from Queue import Empty

import cloudpickle

from ..messages import PythonTask
from ..queue import Queue as ZkQueue
from ..scheduler import AsyncResult, QueueScheduler, Running
from ..utils import timeout as seconds
from ..utils import TimeoutError

__all__ = ('Pool',
           'Queue',
           'AsyncResult')


class Queue(ZkQueue):  # multiprocessing compatible queue

    def __bool__(self):
        return True

    def __nonzero__(self):
        return True

    def get(self, block=True, timeout=-1):
        result = super(Queue, self).get()

        if block:
            try:
                with seconds(timeout):
                    while result is None:
                        result = super(Queue, self).get()
                        time.sleep(0.1)
            except TimeoutError:
                raise Empty

        return cloudpickle.loads(result)

    def put(self, item):
        value = cloudpickle.dumps(item)
        return super(Queue, self).put(value)

    def qsize(self):
        return len(self)

    def empty(self):
        return len(self) == 0


class Pool(Running):

    def __init__(self, processes=-1, *args, **kwargs):
        self.processes = processes
        self.scheduler = QueueScheduler()
        super(Pool, self).__init__(self.scheduler, *args, **kwargs)

    def close(self):
        self.stop()

    def terminate(self):
        self.stop()

    def join(self):
        self.join()

    def wait(self, seconds=-1):
        self.scheduler.wait(seconds)

    def map(self, func, iterable, chunksize=1, **kwargs):
        results = self.map_async(func, iterable, chunksize, **kwargs)
        return [result.get(timeout=-1) for result in results]

    def map_async(self, func, iterable, chunksize=1, callback=None, **kwargs):
        return [self.apply_async(func, (item,), **kwargs) for item in iterable]

    def apply(self, func, args=[], kwds={}, **kwargs):
        result = self.apply_async(func=func, args=args, kwds=kwds, **kwargs)
        return result.get(timeout=-1)

    def apply_async(self, func, args=[], kwds={}, callback=None, **kwargs):
        # TODO: callback
        task = PythonTask(name='multiprocessing-task',
                          fn=func, args=args, kwargs=kwds, **kwargs)

        return self.scheduler.submit(task)



from __future__ import absolute_import, division, print_function

import atexit
import logging
import os
import signal
import time
from collections import Counter

from mesos.interface import mesos_pb2
from mesos.native import MesosSchedulerDriver

from .binpack import bfd
from .interface import Scheduler
from .proxies import SchedulerProxy
from .proxies.messages import FrameworkInfo, TaskInfo, encode
from .utils import timeout


class Running(object):

    def __init__(self, scheduler, name, user='', master=os.getenv('MESOS_MASTER'),
                 implicit_acknowledge=1, *args, **kwargs):
        scheduler = SchedulerProxy(scheduler)
        framework = FrameworkInfo(name=name, user=user, *args, **kwargs)
        self.driver = MesosSchedulerDriver(scheduler, encode(framework),
                                           master, implicit_acknowledge)

        def shutdown(signal, frame):
            self.driver.stop()
        signal.signal(signal.SIGINT, shutdown)
        signal.signal(signal.SIGTERM, shutdown)
        atexit.register(self.driver.stop)

    def run(self):
        return self.driver.run()

    def start(self):
        status = self.driver.start()
        assert status == mesos_pb2.DRIVER_RUNNING
        return status

    def stop(self):
        logging.info("Stopping Mesos driver")
        self.driver.stop()
        logging.info("Joining Mesos driver")
        result = self.driver.join()
        logging.info("Joined Mesos driver")
        if result != mesos_pb2.DRIVER_STOPPED:
            raise RuntimeError("Mesos driver failed with %i", result)

    def join(self):
        return self.driver.join()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()


class AsyncResult(object):

    def __init__(self):
        self.success = None
        self.value = None

    def get(self, timeout=60):
        self.wait(timeout)
        if self.successful():
            return self.value
        else:
            try:
                raise self.value
            except TypeError:
                raise ValueError('Async result indicate task failed!')

    def wait(self, seconds=60):
        with timeout(seconds):
            while not self.ready():
                time.sleep(0.1)

    def ready(self):
        return self.success is not None

    def successful(self):
        return self.success is True


class QueueScheduler(Scheduler):

    def __init__(self, *args, **kwargs):
        self.tasks = {}    # holding task_id => (task, status, result) pairs
        self.results = {}
        self.statuses = {}

    def is_idle(self):
        return not len(self.tasks)

    def report(self):
        states = [status.state for status in self.statuses.values()]
        counts = Counter(states)
        message = ', '.join(['{}: {}'.format(key, count)
                             for key, count in counts.items()])
        logging.info('Task states: {}'.format(message))

    def wait(self, seconds=-1):
        with timeout(seconds):
            while not self.is_idle():
                time.sleep(0.1)

    def submit(self, task):  # supports commandtask, pythontask etc.
        assert isinstance(task, TaskInfo)
        self.tasks[task.id] = task
        self.statuses[task.id] = task.status('TASK_STAGING')
        self.results[task.id] = AsyncResult()
        return self.results[task.id]

    def on_offers(self, driver, offers):
        logging.info('Received offers: {}'.format(sum(offers)))
        self.report()

        # maybe limit to the first n tasks
        staging = [self.tasks[status.task_id]
                   for status in self.statuses.values() if status.is_staging()]

        # best-fit-decreasing binpacking
        bins, skip = bfd(staging, offers, cpus=1, mem=1)

        for offer, tasks in bins:
            try:
                for task in tasks:
                    task.slave_id = offer.slave_id
                    self.statuses[task.id] = task.status('TASK_STARTING')

                # running with empty task list will decline the offer
                logging.info('launched tasks: {}'.format(
                    ', '.join(map(str, tasks))))
                driver.launch(offer.id, tasks)

                self.report()
            except Exception:
                logging.exception('Exception occured during task launch!')

    def on_update(self, driver, status):
        self.statuses[status.task_id] = status
        task = self.tasks[status.task_id]

        logging.info('Updated task {} state {}'.format(
            status.task_id, status.state))
        self.report()

        if status.has_terminated():
            result = self.results[task.id]
            result.success = status.has_succeeded()
            result.value = status.data

            del self.tasks[task.id]
            del self.results[task.id]
            del self.statuses[task.id]

        task.update(status)


if __name__ == '__main__':
    scheduler = QueueScheduler()
    with Running(scheduler, name='test') as fw:
        scheduler.wait()



from __future__ import absolute_import, division, print_function


class Scheduler(object):

    def on_registered(self, driver, framework_id, master):
        """Event handler triggered when the scheduler successfully registers
           with a master.

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        framework_id : string
            Unique ID generated by the master
        master : Master
            Information about the master itself
        Returns
        -------
        self
        """
        pass

    def on_reregistered(self, driver, framework_id, master):
        """Event handler triggered when the scheduler re-registers with a newly
           elected master.

        This is only called when the scheduler has previously been registered.
        masterInfo contains information about the newly elected master.

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        framework_id : string
            Unique ID generated by the master
        master : Master
            Information about the master itself
        """
        pass

    def on_disconnected(self, driver):
        """Event handler triggereg when the scheduler becomes disconnected from
           the master.

        (e.g. the master fails and another is taking over)

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        """
        pass

    def on_offers(self, driver, offers):
        """Event handler triggered when resources have been offered to this
           framework.

        A single offer will only contain resources from a single slave.
        Resources associated with an offer will not be re-offered to _this_
        framework until either (a) this framework has rejected those resources
        (see SchedulerDriver.launchTasks) or (b) those resources have been
        rescinded (see Scheduler.on_rescinded).

        Note that resources may be concurrently offered to more than one
        framework at a time (depending on the allocator being used).  In that
        case, the first framework to launch tasks using those resources will be
        able to use them while the other frameworks will have those resources
        rescinded (or if a framework has already launched tasks with those
        resources then those tasks will fail with a TASK_LOST status and a
        message saying as much).

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        offers: list of Offer
            Resource offer instances
        """
        pass

    def on_rescinded(driver, offer_id):
        """Event handler triggered when an offer is no longer valid.

        (e.g., the slave was lost or another framework used resources in the
        offer)

        If for whatever reason an offer is never rescinded (e.g., dropped
        message, failing over framework, etc.), a framework that attempts to
        launch tasks using an invalid offer will receive TASK_LOST status
        updates for those tasks (see Scheduler.on_offers).

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        offer_id: string
            The unique identifier of the Mesos offer
        """
        pass

    def on_update(self, driver, status):
        """Event handler triggered when the status of a task has changed.

        (e.g., a slave is lost and so the task is lost, a task finishes and an
        executor sends a status update saying so, etc.)

        If implicit acknowledgements are being used, then returning from this
        callback _acknowledges_ receipt of this status update!

        If for  whatever reason the scheduler aborts during this callback (or
        the process exits) another status update will be delivered (note,
        however, that this is currently not true if the slave sending the status
        update is lost/fails during that time).

        If explicit acknowledgements are in use, the scheduler must acknowledge
        this status on the driver.

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        offer_id: string
            The unique identifier of the Mesos offer
        """
        pass

    def on_message(self, driver, executor_id, slave_id, message):
        """Event handler triggered when an executor sends a message.

        These messages are best effort; do not expect a framework message to be
        retransmitted in any reliable fashion.

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        executor_id: string
            The unique identifier of the Mesos executor the message came from
        slave_id: string
            The unique identifier of the Mesos slave the message came from
        message: string
            Arbitrary byte stream
        """
        pass

    def on_slave_lost(self, driver, slave_id):
        """Event handler triggered when a slave has been determined unreachable.

        (e.g., machine failure, network partition.)

        Most frameworks will need to reschedule any tasks launched on this slave
        on a new slave.

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        slave_id: string
            The unique identifier of the lost Mesos slave
        """
        pass

    def on_executor_lost(self, driver, executor_id, slave_id, status):
        """Event handler triggered when an executor has exited/terminated.

        Note that any tasks running will have TASK_LOST status updates
        automatically generated.

        NOTE: This callback is not reliably delivered.

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        executor_id: string
            The unique identifier of the lost Mesos executor
        slave_id: string
            The unique identifier of the Mesos slave where the executor loss
            happened
        status: int
            TODO: figure it out
        """
        pass

    def on_error(self, driver, message):
        """Event handler triggered when there is an unrecoverable error in the
           scheduler or scheduler driver.

        The driver will be aborted BEFORE invoking this callback.

        Parameters
        ----------
        driver: SchedulerDriver
            Interface for interacting with Mesos Master
        message: string
            Arbitrary byte stream
        """
        pass


class Executor(object):

    """Base class for Mesos executors.

    Users' executors should extend this class to get default implementations of
    methods they don't override.
    """

    def on_registered(self, driver, executor, framework, slave):
        """Event handler triggered when the executor driver has been able to
           successfully connect with Mesos.

        In particular, a scheduler can pass some data to its executors through
        the FrameworkInfo.ExecutorInfo's data field.

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        executor: Executor
            The unique identifier of the lost Mesos executor
        framework: Framework
            TODO: write docs
        slave: Salve
            TODO: write docs
        """
        pass

    def on_reregistered(self, driver, slave):
        """Event handler triggered when the executor re-registers with a
           restarted slave.

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        slave: Slave
            TODO: write docs
        """
        pass

    def on_disconnected(self, driver):
        """Event handler triggered when the executor becomes "disconnected" from
           the slave.

        (e.g., the slave is being restarted due to an upgrade)

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        """
        pass

    def on_launch(self, driver, task):
        """Event handler triggered when a task has been launched on this
           executor (initiated via Scheduler.launch).

        Note that this task can be realized with a thread, a process, or some
        simple computation, however, no other callbacks will be invoked on this
        executor until this callback has returned.

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        task: Task
            TODO: write docs
        """
        pass

    def on_kill(self, driver, task_id):
        """Event handler triggered when a task running within this executor has
           been killed (via SchedulerDriver.kill).

        Note that no status update will be sent on behalf of the executor, the
        executor is responsible for creating a new TaskStatus (i.e., with
        TASK_KILLED) and invoking ExecutorDriver's update method.

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        task_id: string
            Unique identifier of the killed task
        """
        pass

    def on_message(self, driver, message):
        """Event handler triggered when a framework message has arrived for this
           executor.

        These messages are best effort; do not expect a framework message to be
        retransmitted in any reliable fashion.

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        message: string
            Arbitrary byte stream
        """
        pass

    def on_shutdown(self, driver):
        """Event handler triggered when the executor should terminate all of its
           currently running tasks.

        Note that after Mesos has determined that an executor has terminated any
        tasks that the executor did not send terminal status updates for (e.g.,
        TASK_KILLED, TASK_FINISHED, TASK_FAILED, etc) a TASK_LOST status update
        will be created.

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        """
        pass

    def on_error(self, driver, message):
        """Event handler triggered when a fatal error has occurred with the
           executor and/or executor driver.

        The driver will be aborted BEFORE invoking this callback.

        Parameters
        ----------
        driver: ExecutorDriver
            Interface for interacting with Mesos Slave
        message: string
            Arbitrary byte stream
        """
        pass



from __future__ import absolute_import, division, print_function

import logging
import sys

from mesos.interface import Scheduler

from .messages import Filters, decode, encode


class SchedulerProxy(Scheduler):

    def __init__(self, scheduler):
        self.scheduler = scheduler

    def registered(self, driver, frameworkId, masterInfo):
        logging.info("Registered with master", extra=dict())
        return self.scheduler.on_registered(SchedulerDriverProxy(driver),
                                            decode(frameworkId),
                                            decode(masterInfo))

    def reregistered(self, driver, masterInfo):
        logging.info("Re-registered with master", extra=dict())
        return self.scheduler.on_reregistered(SchedulerDriverProxy(driver),
                                              decode(masterInfo))

    def disconnected(self, driver):
        logging.debug("Disconnected from master")
        return self.scheduler.on_disconnected(SchedulerDriverProxy(driver))

    def resourceOffers(self, driver, offers):
        logging.debug("Got resource offers",
                      extra=dict(num_offers=len(offers)))
        return self.scheduler.on_offers(SchedulerDriverProxy(driver),
                                        map(decode, offers))

    def offerRescinded(self, driver, offerId):
        logging.debug('Offer rescinded', extra=dict(offer_id=offerId))
        return self.scheduler.on_rescinded(SchedulerDriverProxy(driver),
                                           decode(offerId))

    def statusUpdate(self, driver, status):
        logging.debug('Status update received', extra=dict(
            state=status.state, description=status.message))
        return self.scheduler.on_update(SchedulerDriverProxy(driver),
                                        decode(status))

    def frameworkMessage(self, driver, executorId, slaveId, message):
        logging.debug('Framework message received')
        return self.scheduler.on_message(SchedulerDriverProxy(driver),
                                         decode(executorId),
                                         decode(slaveId),
                                         message)

    def slaveLost(self, driver, slaveId):
        logging.debug('Slave has been lost, tasks should be rescheduled')
        return self.scheduler.on_slave_lost(SchedulerDriverProxy(driver),
                                            decode(slaveId))

    def executorLost(self, driver, executorId, slaveId, status):
        logging.debug('Executor has been lost')
        return self.scheduler.on_executor_lost(SchedulerDriverProxy(driver),
                                               decode(executorId),
                                               decode(slaveId),
                                               status)

    def error(self, driver, message):
        print("Error from Mesos: %s " % message, file=sys.stderr)
        return self.scheduler.on_error(SchedulerDriverProxy(driver), message)


class SchedulerDriverProxy(object):
    """Proxy Interface for Mesos scheduler drivers."""

    def __init__(self, driver):
        self.driver = driver

    def start(self):
        """Starts the scheduler driver.

        This needs to be called before any other driver calls are made.
        """
        return self.driver.start()

    def stop(self, failover=False):
        """Stops the scheduler driver.

        If the 'failover' flag is set to False then it is expected that this
        framework will never reconnect to Mesos and all of its executors and
        tasks can be terminated.  Otherwise, all executors and tasks will
        remain running (for some framework specific failover timeout) allowing
        the scheduler to reconnect (possibly in the same process, or from a
        different process, for example, on a different machine.)
        """
        return self.driver.stop(failover)

    def abort(self):
        """Aborts the driver so that no more callbacks can be made to the
           scheduler.

        The semantics of abort and stop have deliberately been separated so that
        code can detect an aborted driver (i.e., via the return status of
        SchedulerDriver.join), and instantiate and start another driver if
        desired (from within the same process.)
        """
        return self.driver.abort()

    def join(self):
        """Waits for the driver to be stopped or aborted, possibly blocking the
           current thread indefinitely.

        The return status of this function can be used to determine if the
        driver was aborted (see mesos.proto for a description of Status).
        """
        return self.driver.join()

    def request(self, requests):
        """Requests resources from Mesos.

        (see mesos.proto for a description of Request and how, for example, to
        request resources from specific slaves.)

        Any resources available are offered to the framework via
        Scheduler.resourceOffers callback, asynchronously.
        """
        return self.driver.requestResources(map(encode, requests))

    def launch(self, offer_id, tasks, filters=Filters()):
        """Launches the given set of tasks.

        Any resources remaining (i.e., not used by the tasks or their executors)
        will be considered declined.
        The specified filters are applied on all unused resources (see
        mesos.proto for a description of Filters). Available resources are
        aggregated when multiple offers are provided. Note that all offers must
        belong to the same slave. Invoking this function with an empty
        collection of tasks declines the offers in entirety (see
        Scheduler.decline).

        Note that passing a single offer is also supported.
        """
        return self.driver.launchTasks(encode(offer_id),
                                       map(encode, tasks),
                                       encode(filters))

    def kill(self, task_id):
        """Kills the specified task.

        Note that attempting to kill a task is currently not reliable.
        If, for example, a scheduler fails over while it was attempting to kill
        a task it will need to retry in the future.
        Likewise, if unregistered / disconnected, the request will be dropped
        (these semantics may be changed in the future).
        """
        return self.driver.killTask(encode(task_id))

    def reconcile(self, statuses):
        """Allows the framework to query the status for non-terminal tasks.

        This causes the master to send back the latest task status for each task
        in 'statuses', if possible. Tasks that are no longer known will result
        in a TASK_LOST update. If statuses is empty, then the master will send
        the latest status for each task currently known.
        """
        return self.driver.reconcileTasks(map(encode, statuses))

    def decline(self, offer_id, filters=Filters()):
        """Declines an offer in its entirety and applies the specified
           filters on the resources (see mesos.proto for a description of
           Filters).

        Note that this can be done at any time, it is not necessary to do this
        within the Scheduler::resourceOffers callback.
        """
        return self.driver.declineOffer(encode(offer_id),
                                        encode(filters))  # TODO filters

    def accept(self, offer_ids, operations, filters=Filters()):
        """Accepts the given offers and performs a sequence of operations
           on those accepted offers.

        See Offer.Operation in mesos.proto for the set of available operations.
        Available resources are aggregated when multiple offers are provided.

        Note that all offers must belong to the same slave. Any unused resources
        will be considered declined. The specified filters are applied on all
        unused resources (see mesos.proto for a description of Filters).
        """
        return self.driver.acceptOffers(map(encode, offer_ids),
                                        map(encode, operations),
                                        encode(filters))

    def revive(self):
        """Removes all filters previously set by the framework (via
           launchTasks()).

        This enables the framework to receive offers from those filtered slaves.
        """
        return self.driver.reviveOffers()

    def suppress(self):
        """Inform Mesos master to stop sending offers to the framework.

        The scheduler should call reviveOffers() to resume getting offers.
        """
        return self.driver.suppressOffers()

    def acknowledge(self, status):
        """Acknowledges the status update.

        This should only be called once the status update is processed durably
        by the scheduler.

        Not that explicit acknowledgements must be requested via the constructor
        argument, otherwise a call to this method will cause the driver to
        crash.
        """
        return self.driver.acknowledgeStatusUpdate(encode(status))

    def message(self, executor_id, slave_id, message):
        """Sends a message from the framework to one of its executors.

        These messages are best effort; do not expect a framework message to be
        retransmitted in any reliable fashion.
        """
        return self.driver.sendFrameworkMessage(encode(executor_id),
                                                encode(slave_id),
                                                message)



from __future__ import absolute_import, division, print_function

import pytest
from mesos.interface import mesos_pb2
from satyr.proxies.messages import (CommandInfo, Cpus, Disk, FrameworkID,
                                    FrameworkInfo, Map, Mem, MessageProxy,
                                    Offer, RegisterProxies, ResourcesMixin,
                                    ScalarResource, TaskID, TaskInfo,
                                    TaskStatus, decode, encode)


@pytest.fixture
def d():
    return {'a': 1,
            'b': [{'j': 9},
                  {'g': 7, 'h': 8}],
            'c': {'d': 4,
                  'e': {'f': 6}}}


def test_map_init(d):
    m = Map(**d)
    assert isinstance(m, Map)
    assert isinstance(m, dict)


def test_map_get(d):
    m = Map(**d)
    assert m['a'] == 1
    assert m['c']['e']['f'] == 6
    assert m['b'][0]['j'] == 9
    assert m['b'][1]['g'] == 7
    assert isinstance(m['b'], list)
    assert isinstance(m['b'][1], Map)


def test_map_dot_get(d):
    m = Map(**d)
    assert m.a == 1
    assert m.c.e.f == 6
    assert m.b[0].j == 9
    assert m.b[1].g == 7
    assert isinstance(m.b, list)
    assert isinstance(m.b[1], Map)


def test_map_set(d):
    m = Map(**d)
    m['A'] = 11
    m['a'] = 'one'
    m['z'] = {'Z': {'omega': 20}}
    assert m['a'] == 'one'
    assert m['A'] == 11
    assert m['z']['Z']['omega'] == 20
    assert isinstance(m['z'], Map)
    assert isinstance(m['z']['Z'], Map)


def test_map_dot_set(d):
    m = Map(**d)
    m.A = 11
    m.a = 'one'
    m.z = {'Z': {'omega': 20}}
    assert m.a == 'one'
    assert m.A == 11
    assert m.z.Z.omega == 20
    assert isinstance(m.z, Map)
    assert isinstance(m.z.Z, Map)


# def test_map_set_missing(d):
#     m = Map(**d)
#     m['y']['o']['w'] = 9
#     m.y.w.o = 6

#     assert m['y']['o']['w'] == 9
#     assert m.y.w.o == 6


def test_hash():
    d1 = Map(a=Map(b=3), c=5)
    d2 = Map(a=Map(b=3), c=5)
    d3 = Map(a=Map(b=6), c=5)

    assert hash(d1) == hash(d2)
    assert hash(d1) != hash(d3)
    assert hash(d2) != hash(d3)


def test_dict_hashing():
    d2 = Map(a=Map(b=3), c=5)
    d3 = Map(a=Map(b=6), c=5)

    c = {}
    c[d2.a] = d2
    c[d3.a] = d3

    assert c[d2.a] == d2
    assert c[d3.a] == d3


def test_register_proxies():
    class Base(object):
        __metaclass__ = RegisterProxies
        proto = 'base'

    class First(Base):
        proto = 'first'

    class Second(Base):
        proto = 'second'

    class Third(Base):
        proto = 'third'

    assert Base.registry == [('third', Third),
                             ('second', Second),
                             ('first', First),
                             ('base', Base)]


def test_encode_resources():
    pb = encode(Cpus(0.1))
    assert pb.scalar.value == 0.1
    assert pb.name == 'cpus'
    assert pb.type == mesos_pb2.Value.SCALAR

    pb = encode(Mem(16))
    assert pb.scalar.value == 16
    assert pb.name == 'mem'
    assert pb.type == mesos_pb2.Value.SCALAR

    pb = encode(Disk(256))
    assert pb.scalar.value == 256
    assert pb.name == 'disk'
    assert pb.type == mesos_pb2.Value.SCALAR


def test_encode_task_info_resources():
    task = TaskInfo(name='test-task',
                    id=TaskID(value='test-task-id'),
                    resources=[Cpus(0.1), Mem(16)],
                    command=CommandInfo(value='testcmd'))
    pb = encode(task)
    assert pb.name == 'test-task'
    assert pb.task_id.value == 'test-task-id'
    assert pb.resources[0].name == 'cpus'
    assert pb.resources[0].scalar.value == 0.1
    assert pb.resources[1].name == 'mem'
    assert pb.resources[1].scalar.value == 16
    assert pb.command.value == 'testcmd'


def test_decode_framework_info():
    message = mesos_pb2.FrameworkInfo(id=mesos_pb2.FrameworkID(value='test'))
    wrapped = decode(message)

    assert isinstance(wrapped, MessageProxy)
    assert isinstance(wrapped, FrameworkInfo)
    assert isinstance(wrapped.id, MessageProxy)
    assert isinstance(wrapped.id, FrameworkID)


def test_scalar_resource_comparison():
    r1 = ScalarResource(value=11.5)

    assert r1 == ScalarResource(value=11.5)
    assert r1 <= ScalarResource(value=11.5)
    assert r1 >= ScalarResource(value=11.5)
    assert r1 < ScalarResource(value=12)
    assert r1 > ScalarResource(value=11)

    assert r1 == 11.5
    assert r1 <= 11.5
    assert r1 >= 11.5
    assert r1 < 12
    assert r1 > 11


def test_scalar_resource_addition():
    r1 = ScalarResource(value=11.5)
    r2 = ScalarResource(value=2)

    s = r1 + r2
    assert isinstance(s, ScalarResource)
    assert s == ScalarResource(13.5)
    assert s == 13.5


def test_scalar_resource_sum():
    r1 = ScalarResource(value=11.5)
    r2 = ScalarResource(value=2)
    r3 = ScalarResource(value=3)

    s = sum([r1, r2, r3])
    assert isinstance(s, ScalarResource)
    assert s == ScalarResource(16.5)
    assert s == 16.5


def test_scalar_resource_subtraction():
    r1 = ScalarResource(value=11.5)
    r2 = ScalarResource(value=2)

    s = r1 - r2
    assert isinstance(s, ScalarResource)
    assert s == ScalarResource(9.5)
    assert s == 9.5


def test_scalar_resource_inplace_addition():
    r1 = ScalarResource(value=11.5)
    r2 = ScalarResource(value=2)

    r1 += r2
    assert isinstance(r1, ScalarResource)
    assert r1 == ScalarResource(13.5)
    assert r1 == 13.5


def test_scalar_resource_inplace_subtraction():
    r1 = ScalarResource(value=11.5)
    r2 = ScalarResource(value=2)

    r1 -= r2
    assert isinstance(r1, ScalarResource)
    assert r1 == ScalarResource(9.5)
    assert r1 == 9.5


def test_scalar_resource_multiplication():
    r1 = ScalarResource(value=11.5)
    r2 = ScalarResource(value=2)

    m = r1 * r2
    assert isinstance(m, ScalarResource)
    assert m == ScalarResource(23)
    assert m == 23


def test_scalar_resource_division():
    r1 = ScalarResource(value=11.5)
    r2 = ScalarResource(value=2)

    d = r1 / r2
    assert isinstance(d, ScalarResource)
    assert d == ScalarResource(5.75)
    assert d == 5.75


def test_resources_mixin_comparison():
    o1 = Offer(resources=[Cpus(1), Mem(128), Disk(0)])
    o2 = Offer(resources=[Cpus(2), Mem(256), Disk(1024)])

    t1 = TaskInfo(resources=[Cpus(0.5), Mem(128), Disk(0)])
    t2 = TaskInfo(resources=[Cpus(1), Mem(256), Disk(512)])
    t3 = TaskInfo(resources=[Cpus(0.5), Mem(256), Disk(512)])

    assert o1.cpus == 1
    assert o1.mem == 128
    assert o2.cpus == 2
    assert o2.disk == 1024

    assert t1.cpus == 0.5
    assert t1.mem == 128
    assert t2.cpus == 1
    assert t2.disk == 512

    assert o1 == o1
    assert o1 < o2
    assert o1 <= o2
    assert o2 > o1
    assert o2 >= o1

    assert t1 == t1
    assert t1 < t2
    assert t1 <= t2
    assert t2 > t1
    assert t2 >= t1

    assert o1 >= t1
    assert o2 >= t1
    assert o2 >= t2
    assert t2 >= o1

    assert t3 > o1
    assert t3 <= t2
    assert t3 > t1


def test_resources_mixin_addition():
    o = Offer(resources=[Cpus(1), Mem(128), Disk(0)])
    t = TaskInfo(resources=[Cpus(0.5), Mem(128), Disk(0)])

    s = o + t
    assert isinstance(s, ResourcesMixin)
    assert s.cpus == Cpus(1.5)
    assert s.cpus == 1.5
    assert s.mem == Mem(256)
    assert s.mem == 256
    assert s.disk == Disk(0)
    assert s.disk == 0


def test_resources_mixin_sum():
    o1 = Offer(resources=[Cpus(1), Mem(128), Disk(0)])
    o2 = Offer(resources=[Cpus(2), Mem(128), Disk(100)])
    o3 = Offer(resources=[Cpus(0.5), Mem(256), Disk(200)])

    s = sum([o1, o2, o3])
    assert isinstance(s, ResourcesMixin)
    assert s.cpus == Cpus(3.5)
    assert s.cpus == 3.5
    assert s.mem == Mem(512)
    assert s.mem == 512
    assert s.disk == Disk(300)
    assert s.disk == 300


def test_resources_mixin_subtraction():
    o = Offer(resources=[Cpus(1), Mem(128), Disk(0)])
    t = TaskInfo(resources=[Cpus(0.5), Mem(128), Disk(0)])

    s = o - t
    assert isinstance(s, ResourcesMixin)
    assert s.cpus == Cpus(0.5)
    assert s.cpus == 0.5
    assert s.mem == Mem(0)
    assert s.mem == 0
    assert s.disk == Disk(0)
    assert s.disk == 0


def test_resources_mixin_inplace_addition():
    o = Offer(resources=[Cpus(1), Mem(128), Disk(64)])
    t = TaskInfo(resources=[Cpus(0.5), Mem(128), Disk(0)])

    o += t
    assert isinstance(o, Offer)
    assert o.cpus == Cpus(1.5)
    assert o.cpus == 1.5
    assert o.mem == Mem(256)
    assert o.mem == 256
    assert o.disk == Disk(64)
    assert o.disk == 64


def test_resources_mixin_inplace_subtraction():
    o = Offer(resources=[Cpus(1), Mem(128), Disk(64)])
    t = TaskInfo(resources=[Cpus(0.5), Mem(128), Disk(0)])

    o -= t
    assert isinstance(o, Offer)
    assert o.cpus == Cpus(0.5)
    assert o.cpus == 0.5
    assert o.mem == Mem(0)
    assert o.mem == 0
    assert o.disk == Disk(64)
    assert o.disk == 64


def test_encode_task_info():
    t = TaskInfo(name='test-task',
                 id=TaskID(value='test-task-id'),
                 resources=[Cpus(0.1), Mem(16)],
                 command=CommandInfo(value='echo 100'))

    p = encode(t)
    assert isinstance(p, mesos_pb2.TaskInfo)
    assert p.command.value == 'echo 100'
    assert p.name == 'test-task'
    assert p.resources[0].name == 'cpus'
    assert p.resources[0].scalar.value == 0.1
    assert p.task_id.value == 'test-task-id'


def test_non_strict_encode_task_info():
    t = TaskInfo(name='test-task',
                 id=TaskID(value='test-task-id'),
                 resources=[Cpus(0.1), Mem(16)],
                 command=CommandInfo(value='echo 100'))
    t.result = 'some binary data'
    t.status = TaskStatus()

    p = encode(t)
    assert isinstance(p, mesos_pb2.TaskInfo)
    assert p.command.value == 'echo 100'
    with pytest.raises(AttributeError):
        p.status



from mesos.interface import mesos_pb2
from satyr.proxies import ExecutorDriverProxy, ExecutorProxy
from satyr.proxies.messages import TaskStatus


def test_executor_event_handlers(mocker):
    executor = mocker.Mock()
    driver = mocker.Mock()
    proxy = ExecutorProxy(executor)

    proxy.registered(driver, mesos_pb2.ExecutorInfo(), mesos_pb2.FrameworkInfo(),
                     mesos_pb2.SlaveInfo())
    proxy.reregistered(driver, mesos_pb2.SlaveInfo())
    proxy.disconnected(driver)
    proxy.launchTask(driver, mesos_pb2.TaskInfo())
    proxy.killTask(driver, mesos_pb2.TaskID())
    proxy.frameworkMessage(driver, 'message')
    proxy.shutdown(driver)
    proxy.error(driver, 'message')

    executor.on_registered.assert_called_once()
    executor.on_reregistered.assert_called_once()
    executor.on_disconnected.assert_called_once()
    executor.on_launch.assert_called_once()
    executor.on_kill.assert_called_once()
    executor.on_message.assert_called_once()
    executor.on_shutdown.assert_called_once()
    executor.on_error.assert_called_once()


def test_executor_driver_callbacks(mocker):
    driver = mocker.Mock()
    proxy = ExecutorDriverProxy(driver)

    proxy.abort()
    proxy.join()
    proxy.start()
    proxy.stop()
    proxy.run()
    proxy.update(TaskStatus())
    proxy.message('message')

    driver.abort.assert_called_once()
    driver.join.assert_called_once()
    driver.start.assert_called_once()
    driver.stop.assert_called_once()
    driver.run.assert_called_once()
    driver.sendStatusUpdate.assert_called_once()
    driver.sendFrameworkMessage.assert_called_once()



from __future__ import absolute_import, division, print_function

import pytest
from satyr.messages import PythonTask
from satyr.proxies.messages import (CommandInfo, ContainerInfo, Cpus, Disk,
                                    DockerInfo, Mem, TaskID, TaskInfo)
from satyr.scheduler import QueueScheduler, Running


@pytest.fixture
def command():
    task = TaskInfo(name='test-task',
                    id=TaskID(value='test-task-id'),
                    resources=[Cpus(0.1), Mem(64)],
                    command=CommandInfo(value='echo 100'))
    return task


@pytest.fixture
def docker_command():
    task = TaskInfo(name='test-docker-task',
                    id=TaskID(value='test-docker-task-id'),
                    resources=[Cpus(0.1), Mem(64)],
                    command=CommandInfo(value='echo 100'),
                    container=ContainerInfo(
                        type='DOCKER',
                        docker=DockerInfo(image='lensa/satyr:latest')))
    return task


@pytest.fixture
def docker_python():
    task = PythonTask(id=TaskID(value='test-python-task-id'),
                      fn=sum, args=[range(5)],
                      name='test-python-task-name',
                      resources=[Cpus(0.1), Mem(64), Disk(0)])
    return task


def test_command(mocker, command):
    sched = QueueScheduler()
    mocker.spy(sched, 'on_update')

    with Running(sched, name='test-scheduler'):
        sched.submit(command)
        sched.wait()  # block until all tasks finishes

    calls = sched.on_update.call_args_list
    assert len(calls) == 2

    args, kwargs = calls[0]
    assert args[1].task_id.value == 'test-task-id'
    assert args[1].state == 'TASK_RUNNING'

    args, kwargs = calls[1]
    assert args[1].task_id.value == 'test-task-id'
    assert args[1].state == 'TASK_FINISHED'


def test_docker_command(mocker, docker_command):
    sched = QueueScheduler()
    mocker.spy(sched, 'on_update')

    with Running(sched, name='test-scheduler'):
        sched.submit(docker_command)
        sched.wait()  # block until all tasks finishes

    calls = sched.on_update.call_args_list
    assert len(calls) == 2

    args, kwargs = calls[0]
    assert args[1].task_id.value == 'test-docker-task-id'
    assert args[1].state == 'TASK_RUNNING'

    args, kwargs = calls[1]
    assert args[1].task_id.value == 'test-docker-task-id'
    assert args[1].state == 'TASK_FINISHED'


def test_docker_python(mocker, docker_python):
    sched = QueueScheduler()
    mocker.spy(sched, 'on_update')

    with Running(sched, name='test-scheduler'):
        sched.submit(docker_python)
        sched.wait()  # block until all tasks finishes

    calls = sched.on_update.call_args_list
    assert len(calls) == 2

    args, kwargs = calls[0]
    assert args[1].task_id.value == 'test-python-task-id'
    assert args[1].state == 'TASK_RUNNING'

    args, kwargs = calls[1]
    assert args[1].task_id.value == 'test-python-task-id'
    assert args[1].state == 'TASK_FINISHED'


def test_docker_python_exception():
    sched = QueueScheduler()

    def error():
        raise Exception('Dummy exception on executor side!')

    task = PythonTask(id=TaskID(value='test-python-task-id'),
                      fn=error, name='test-python-task-name',
                      resources=[Cpus(0.1), Mem(64), Disk(0)])

    with Running(sched, name='test-scheduler'):
        result = sched.submit(task)
        with pytest.raises(Exception) as e:
            result.get()
        assert e.value.message == 'Dummy exception on executor side!'


def test_multiple_submissions(mocker, docker_python):
    sched = QueueScheduler()
    mocker.spy(sched, 'on_update')

    with Running(sched, name='test-scheduler'):
        results = [None] * 3
        for i in range(3):
            task = PythonTask(id=TaskID(value='test-python-task-{}'.format(i)),
                              fn=sum, args=[[1, 10, i]],
                              name='test-python-task-name',
                              resources=[Cpus(0.1), Mem(64), Disk(0)])
            results[i] = sched.submit(task)
        sched.wait()  # block until all tasks finishes

    assert [r.get() for r in results] == [11, 12, 13]


def test_sequential_submit_get(mocker, docker_python):
    sched = QueueScheduler()
    mocker.spy(sched, 'on_update')

    with Running(sched, name='test-scheduler'):
        for i in range(3):
            task = PythonTask(id=TaskID(value='test-python-task-{}'.format(i)),
                              fn=sum, args=[[1, 10, i]],
                              name='test-python-task-name',
                              resources=[Cpus(0.1), Mem(64), Disk(0)])
            result = sched.submit(task)
            assert result.get(timeout=10) == 11 + i


def test_docker_python_result(mocker, docker_python):
    sched = QueueScheduler()
    with Running(sched, name='test-scheduler'):
        result = sched.submit(docker_python)
        sched.wait()  # block until all tasks finishes
        assert result.get() == 10



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sample.proto

import sys

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2

_b = sys.version_info[0] < 3 and (
    lambda x: x) or (lambda x: x.encode('latin1'))
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name='sample.proto',
    package='tests',
    serialized_pb=_b('\n\x0csample.proto\x12\x05tests\"\xeb\x03\n\x0eMessageOfTypes\x12\x0c\n\x04\x64ubl\x18\x01 \x02(\x01\x12\x0c\n\x04\x66lot\x18\x02 \x02(\x02\x12\x0b\n\x03i32\x18\x03 \x02(\x05\x12\x0b\n\x03i64\x18\x04 \x02(\x03\x12\x0c\n\x04ui32\x18\x05 \x02(\r\x12\x0c\n\x04ui64\x18\x06 \x02(\x04\x12\x0c\n\x04si32\x18\x07 \x02(\x11\x12\x0c\n\x04si64\x18\x08 \x02(\x12\x12\x0b\n\x03\x66\x33\x32\x18\t \x02(\x07\x12\x0b\n\x03\x66\x36\x34\x18\x11 \x02(\x06\x12\x0c\n\x04sf32\x18\n \x02(\x0f\x12\x0c\n\x04sf64\x18\x0b \x02(\x10\x12\x0b\n\x03\x62ol\x18\x0c \x02(\x08\x12\r\n\x05strng\x18\r \x02(\t\x12\x0c\n\x04\x62yts\x18\x0e \x02(\x0c\x12\x30\n\x06nested\x18\x0f \x02(\x0b\x32 .tests.MessageOfTypes.NestedType\x12\'\n\x03\x65nm\x18\x10 \x02(\x0e\x32\x1a.tests.MessageOfTypes.Enum\x12\r\n\x05range\x18\x12 \x03(\x05\x12\x38\n\x0enestedRepeated\x18\x13 \x03(\x0b\x32 .tests.MessageOfTypes.NestedType\x12/\n\x0b\x65nmRepeated\x18\x14 \x03(\x0e\x32\x1a.tests.MessageOfTypes.Enum\x1a\x19\n\nNestedType\x12\x0b\n\x03req\x18\x01 \x02(\t\"\x1b\n\x04\x45num\x12\x05\n\x01\x41\x10\x00\x12\x05\n\x01\x42\x10\x01\x12\x05\n\x01\x43\x10\x02')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


_MESSAGEOFTYPES_ENUM = _descriptor.EnumDescriptor(
    name='Enum',
    full_name='tests.MessageOfTypes.Enum',
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name='A', index=0, number=0,
            options=None,
            type=None),
        _descriptor.EnumValueDescriptor(
            name='B', index=1, number=1,
            options=None,
            type=None),
        _descriptor.EnumValueDescriptor(
            name='C', index=2, number=2,
            options=None,
            type=None),
    ],
    containing_type=None,
    options=None,
    serialized_start=488,
    serialized_end=515,
)
_sym_db.RegisterEnumDescriptor(_MESSAGEOFTYPES_ENUM)


_MESSAGEOFTYPES_NESTEDTYPE = _descriptor.Descriptor(
    name='NestedType',
    full_name='tests.MessageOfTypes.NestedType',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='req', full_name='tests.MessageOfTypes.NestedType.req', index=0,
            number=1, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=461,
    serialized_end=486,
)

_MESSAGEOFTYPES = _descriptor.Descriptor(
    name='MessageOfTypes',
    full_name='tests.MessageOfTypes',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='dubl', full_name='tests.MessageOfTypes.dubl', index=0,
            number=1, type=1, cpp_type=5, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='flot', full_name='tests.MessageOfTypes.flot', index=1,
            number=2, type=2, cpp_type=6, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='i32', full_name='tests.MessageOfTypes.i32', index=2,
            number=3, type=5, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='i64', full_name='tests.MessageOfTypes.i64', index=3,
            number=4, type=3, cpp_type=2, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='ui32', full_name='tests.MessageOfTypes.ui32', index=4,
            number=5, type=13, cpp_type=3, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='ui64', full_name='tests.MessageOfTypes.ui64', index=5,
            number=6, type=4, cpp_type=4, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='si32', full_name='tests.MessageOfTypes.si32', index=6,
            number=7, type=17, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='si64', full_name='tests.MessageOfTypes.si64', index=7,
            number=8, type=18, cpp_type=2, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='f32', full_name='tests.MessageOfTypes.f32', index=8,
            number=9, type=7, cpp_type=3, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='f64', full_name='tests.MessageOfTypes.f64', index=9,
            number=17, type=6, cpp_type=4, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='sf32', full_name='tests.MessageOfTypes.sf32', index=10,
            number=10, type=15, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='sf64', full_name='tests.MessageOfTypes.sf64', index=11,
            number=11, type=16, cpp_type=2, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='bol', full_name='tests.MessageOfTypes.bol', index=12,
            number=12, type=8, cpp_type=7, label=2,
            has_default_value=False, default_value=False,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='strng', full_name='tests.MessageOfTypes.strng', index=13,
            number=13, type=9, cpp_type=9, label=2,
            has_default_value=False, default_value=_b("").decode('utf-8'),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='byts', full_name='tests.MessageOfTypes.byts', index=14,
            number=14, type=12, cpp_type=9, label=2,
            has_default_value=False, default_value=_b(""),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='nested', full_name='tests.MessageOfTypes.nested', index=15,
            number=15, type=11, cpp_type=10, label=2,
            has_default_value=False, default_value=None,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='enm', full_name='tests.MessageOfTypes.enm', index=16,
            number=16, type=14, cpp_type=8, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='range', full_name='tests.MessageOfTypes.range', index=17,
            number=18, type=5, cpp_type=1, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='nestedRepeated', full_name='tests.MessageOfTypes.nestedRepeated', index=18,
            number=19, type=11, cpp_type=10, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='enmRepeated', full_name='tests.MessageOfTypes.enmRepeated', index=19,
            number=20, type=14, cpp_type=8, label=3,
            has_default_value=False, default_value=[],
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[_MESSAGEOFTYPES_NESTEDTYPE, ],
    enum_types=[
        _MESSAGEOFTYPES_ENUM,
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=24,
    serialized_end=515,
)

_MESSAGEOFTYPES_NESTEDTYPE.containing_type = _MESSAGEOFTYPES
_MESSAGEOFTYPES.fields_by_name[
    'nested'].message_type = _MESSAGEOFTYPES_NESTEDTYPE
_MESSAGEOFTYPES.fields_by_name['enm'].enum_type = _MESSAGEOFTYPES_ENUM
_MESSAGEOFTYPES.fields_by_name[
    'nestedRepeated'].message_type = _MESSAGEOFTYPES_NESTEDTYPE
_MESSAGEOFTYPES.fields_by_name['enmRepeated'].enum_type = _MESSAGEOFTYPES_ENUM
_MESSAGEOFTYPES_ENUM.containing_type = _MESSAGEOFTYPES
DESCRIPTOR.message_types_by_name['MessageOfTypes'] = _MESSAGEOFTYPES

MessageOfTypes = _reflection.GeneratedProtocolMessageType('MessageOfTypes', (_message.Message,), dict(

    NestedType=_reflection.GeneratedProtocolMessageType('NestedType', (_message.Message,), dict(
        DESCRIPTOR=_MESSAGEOFTYPES_NESTEDTYPE,
        __module__='sample_pb2'
        # @@protoc_insertion_point(class_scope:tests.MessageOfTypes.NestedType)
    )),
    DESCRIPTOR=_MESSAGEOFTYPES,
    __module__='sample_pb2'
    # @@protoc_insertion_point(class_scope:tests.MessageOfTypes)
))
_sym_db.RegisterMessage(MessageOfTypes)
_sym_db.RegisterMessage(MessageOfTypes.NestedType)


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# coding: utf-8

from os.path import exists

from setuptools import setup

setup(name='satyr',
      version='0.1.3',
      description='Extensible Python Framework for Apache Mesos',
      url='http://github.com/lensacom/satyr',
      maintainer='Krisztián Szűcs',
      maintainer_email='krisztian.szucs@lensa.com',
      license='Apache License, Version 2.0',
      keywords='mesos framework multiprocessing',
      packages=['satyr', 'satyr.proxies', 'satyr.apis'],
      long_description=(open('README.md').read() if exists('README.md')
                        else ''),
      install_requires=['cloudpickle', 'kazoo'],
      extras_require={'mesos': ['mesos.native']},
      setup_requires=['pytest-runner'],
      tests_require=['pytest-mock', 'pytest'],
      zip_safe=False)



from __future__ import absolute_import, division, print_function

import logging

import cloudpickle
from mesos.interface import mesos_pb2

from .proxies.messages import (CommandInfo, ContainerInfo, Cpus, Disk,
                               DockerInfo, ExecutorID, ExecutorInfo, Mem,
                               TaskInfo, TaskStatus)


class PickleMixin(object):

    @property
    def data(self):
        return cloudpickle.loads(self['data'])

    @data.setter
    def data(self, value):
        self['data'] = cloudpickle.dumps(value)


class PythonTaskStatus(PickleMixin, TaskStatus):

    proto = mesos_pb2.TaskStatus(
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))

    def __init__(self, data=None, **kwargs):
        super(PythonTaskStatus, self).__init__(**kwargs)
        self.data = data


class PythonTask(PickleMixin, TaskInfo):

    proto = mesos_pb2.TaskInfo(
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))

    def __init__(self, fn=None, args=[], kwargs={},
                 docker='lensa/satyr:latest', command='python -m satyr.executor',
                 resources=[Cpus(0.1), Mem(128)], **kwds):
        super(PythonTask, self).__init__(**kwds)
        self.executor = ExecutorInfo(
            executor_id=ExecutorID(value=self.id.value),
            command=CommandInfo(value=command, shell=True),
            container=ContainerInfo(type='DOCKER', docker=DockerInfo(
                image=docker, force_pull_image=False, network='HOST')))
        self.resources = resources
        self.data = (fn, args, kwargs)

    def __call__(self):
        fn, args, kwargs = self.data
        return fn(*args, **kwargs)

    def status(self, state, **kwargs):
        return PythonTaskStatus(task_id=self.task_id, state=state, **kwargs)

    def update(self, status):
        self.on_update(status)

        if status.has_succeeded():
            self.on_success(status)
        elif status.has_failed():
            self.on_fail(status)

    def on_update(self, status):
        pass

    def on_success(self, status):
        logging.info('Task {} has been succeded'.format(self.id.value))

    def on_fail(self, status):
        logging.info('Task {} has been failed due to {}'.format(self.id.value,
                                                                status.message))



from mesos.interface import mesos_pb2
from satyr.proxies import SchedulerDriverProxy, SchedulerProxy
from satyr.proxies.messages import (ExecutorID, OfferID, Operation, Request,
                                    SlaveID, TaskInfo, TaskStatus)


def test_scheduler_event_handlers(mocker):
    sched = mocker.Mock()
    driver = mocker.Mock()
    proxy = SchedulerProxy(sched)

    proxy.executorLost(driver, mesos_pb2.ExecutorInfo(),
                       mesos_pb2.ExecutorID(), 1)
    proxy.frameworkMessage(driver, mesos_pb2.ExecutorID(),
                           mesos_pb2.SlaveID(), 'message')
    proxy.offerRescinded(driver, mesos_pb2.OfferID())
    proxy.registered(driver, mesos_pb2.FrameworkID(), mesos_pb2.MasterInfo())
    proxy.resourceOffers(driver, [mesos_pb2.Offer(), mesos_pb2.Offer()])
    proxy.slaveLost(driver, mesos_pb2.SlaveID())
    proxy.statusUpdate(driver, mesos_pb2.TaskStatus())
    proxy.reregistered(driver, mesos_pb2.MasterInfo())
    proxy.error(driver, 'message')
    proxy.disconnected(driver)

    sched.on_executor_lost.assert_called_once()
    sched.on_message.assert_called_once()
    sched.on_rescinded.assert_called_once()
    sched.on_registered.assert_called_once()
    sched.on_offers.assert_called_once()
    sched.on_slave_lost.assert_called_once()
    sched.on_update.assert_called_once()
    sched.on_reregistered.assert_called_once()
    sched.on_error.assert_called_once()
    sched.on_disconnected.assert_called_once()


def test_scheduler_driver_callbacks(mocker):
    driver = mocker.Mock()
    proxy = SchedulerDriverProxy(driver)

    proxy.abort()
    proxy.join()
    proxy.start()
    proxy.stop()
    proxy.acknowledge(TaskStatus())
    proxy.decline(OfferID())
    proxy.kill(OfferID())
    proxy.launch(OfferID(), [TaskInfo()])
    proxy.message(ExecutorID(), SlaveID(), 'message')
    proxy.reconcile([TaskStatus()])
    proxy.request([Request()])
    proxy.revive()
    proxy.suppress()
    proxy.accept(OfferID(), [Operation()])

    driver.abort.assert_called_once()
    driver.join.assert_called_once()
    driver.start.assert_called_once()
    driver.stop.assert_called_once()
    driver.acknowledgeStatusUpdate.assert_called_once()
    driver.declineOffer.assert_called_once()
    driver.killTask.assert_called_once()
    driver.launchTasks.assert_called_once()
    driver.sendFrameworkMessage.assert_called_once()
    driver.reconcileTasks.assert_called_once()
    driver.requestResources.assert_called_once()
    driver.reviveOffers.assert_called_once()
    driver.suppressOffers.assert_called_once()
    driver.acceptOffers.assert_called_once()



import cloudpickle as cp
import pytest
from satyr.queue import LockingQueue, Queue


def test_queue_put_get(zk):
    queue = Queue(zk, '/satyr/putget')
    queue.put(cp.dumps(range(5)))
    assert cp.loads(queue.get()) == range(5)


def test_locking_queue_put_get(zk):
    queue = LockingQueue(zk, '/satyr/putget_locking')
    queue.put(cp.dumps(range(5)))
    assert queue.get() == cp.dumps(range(5))
    queue.consume()


def test_queue_serde(zk):
    queue = Queue(zk, '/satyr/serde')
    queue.put(cp.dumps({'a': 1, 'b': 2}))
    queue.put(cp.dumps({'c': 3}))

    pickled_queue = cp.dumps(queue)
    unpickled_queue = cp.loads(pickled_queue)

    assert cp.loads(unpickled_queue.get()) == {'a': 1, 'b': 2}
    assert cp.loads(unpickled_queue.get()) == {'c': 3}


def test_locking_queue_serde(zk):
    queue = LockingQueue(zk, '/satyr/serde_locking')
    queue.put(cp.dumps({'a': 1, 'b': 2}))
    queue.put(cp.dumps({'c': 3}))

    pickled_queue = cp.dumps(queue)
    unpickled_queue = cp.loads(pickled_queue)

    assert cp.loads(unpickled_queue.get()) == {'a': 1, 'b': 2}
    unpickled_queue.consume()
    assert cp.loads(unpickled_queue.get()) == {'c': 3}
    unpickled_queue.consume()



import threading
import time
from Queue import Empty

import cloudpickle as cp
import pytest
from satyr.apis.multiprocessing import AsyncResult, Pool, Queue
from satyr.proxies.messages import Cpus, Disk, Mem
from satyr.utils import TimeoutError


@pytest.fixture
def resources():
    return [Cpus(0.1), Mem(128), Disk(0)]


def test_queue_size(zk):
    queue = Queue(zk, '/satyr/size')
    assert queue.empty()
    assert queue.qsize() == 0

    queue.put(cp.dumps(range(5)))
    assert queue.empty() is False
    assert queue.qsize() == 1


def test_queue_blocking_get(zk):
    queue = Queue(zk, '/satyr/blocking')

    def delayed_put():
        import time
        time.sleep(2)
        queue.put(cp.dumps(range(5)))

    t = threading.Thread(target=delayed_put)
    t.start()
    with pytest.raises(Empty):
        queue.get(block=True, timeout=1)

    assert queue.get(block=True, timeout=2) == cp.dumps(range(5))


def test_apply_async():
    with Pool(name='test-pool') as pool:
        res1 = pool.apply_async(lambda a, b: a + b, [1, 2])
        res2 = pool.apply_async(lambda a, b: a * b, [3, 5])
        pool.wait()

        assert isinstance(res1, AsyncResult)
        assert isinstance(res2, AsyncResult)

        assert res1.get(timeout=10) == 3
        assert res2.get(timeout=10) == 15


def test_apply_async_timeout():
    with pytest.raises(TimeoutError):
        with Pool(name='test-pool') as pool:
            res = pool.apply_async(time.sleep, (3,))
            res.get(timeout=1)


def test_apply():
    with Pool(name='test-pool') as pool:
        result = pool.apply(lambda a, b: a + b, [1, 3])
        assert result == 4

        result = pool.apply(lambda a, b: a + b, ['a', 'b'])
        assert result == 'ab'
        pool.wait(seconds=20)


def test_multiple_apply_async(resources):
    with Pool(name='test-pool') as pool:
        results = [pool.apply_async(lambda a, b: a + b, [1, i],
                                    resources=resources)
                   for i in range(5)]
        values = [res.get(timeout=20) for res in results]

    assert values == [i + 1 for i in range(5)]


def test_queue_apply_async(zk, resources):
    def feed(i, queue):
        queue.put(cp.dumps(i))

    queue = Queue(zk, '/satyr/test-pool')
    with Pool(name='test-pool') as pool:
        results = [pool.apply_async(feed, [i, queue], resources=resources)
                   for i in range(5)]
        pool.wait(seconds=30)

    time.sleep(1)
    results = [cp.loads(queue.get()) for i in range(5)]
    assert sorted(results) == range(5)


def test_map_async(resources):
    with Pool(name='test-pool') as pool:
        results = pool.map_async(
            lambda tpl: tpl[0] + tpl[1], zip(range(5), range(5)),
            resources=resources)
        assert all([isinstance(res, AsyncResult) for res in results])
        values = [res.get(timeout=30) for res in results]

    assert values == [i + i for i in range(5)]


def test_map(resources):
    with Pool(name='test-pool') as pool:
        results = pool.map(
            lambda tpl: tpl[0] + tpl[1], zip(range(3), range(3)),
            resources=resources)

    expected = [i + i for i in range(3)]
    assert results == expected



from __future__ import absolute_import, division, print_function

import logging
import pkg_resources as _pkg_resources

from .scheduler import QueueScheduler
from .executor import OneOffExecutor
from .messages import PythonTask, PythonTaskStatus  # important to register classes


logging.basicConfig(level=logging.DEBUG,
                    format='%(relativeCreated)6d %(threadName)s %(message)s')

__version__ = _pkg_resources.get_distribution('satyr').version

__all__ = ('QueueScheduler',
           'OneOffExecutor',
           'PythonTask',
           'PythonTaskStatus')



from __future__ import absolute_import, division, print_function

import cloudpickle
from mesos.interface import mesos_pb2
from satyr.messages import PythonTask, PythonTaskStatus
from satyr.proxies.messages import TaskID, decode, encode


def test_python_task_status_decode():
    data = {'arbitrary': 'data', 'lst': [1, 2, 3]}
    dumped = cloudpickle.dumps(data)

    proto = mesos_pb2.TaskStatus(
        data=dumped,
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))
    status = decode(proto)

    assert isinstance(status, PythonTaskStatus)
    assert status['data'] == dumped
    assert status.data == data

    proto = mesos_pb2.TaskStatus(
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))
    status = decode(proto)
    status.data = data

    assert isinstance(status, PythonTaskStatus)
    assert status.data == data
    assert status['data'] == dumped


def test_python_task_status_encode():
    data = {'arbitrary': 'data', 'value': 5}
    dumped = cloudpickle.dumps(data)

    status = PythonTaskStatus(task_id={'value': 'test-id'},
                              state='TASK_STAGING',
                              data=data)

    proto = encode(status)
    assert isinstance(proto, mesos_pb2.TaskStatus)
    assert proto.data == dumped
    assert proto.task_id.value == 'test-id'
    assert proto.state == mesos_pb2.TASK_STAGING

    status = PythonTaskStatus(task_id={'value': 'test-id'},
                              state='TASK_RUNNING')
    status.data = data
    proto = encode(status)
    assert isinstance(proto, mesos_pb2.TaskStatus)
    assert proto.data == dumped
    assert proto.task_id.value == 'test-id'
    assert proto.state == mesos_pb2.TASK_RUNNING


def test_python_task_decode():
    fn, args, kwargs = sum, [range(5)], {}
    data = (fn, args, kwargs)
    dumped = cloudpickle.dumps(data)

    proto = mesos_pb2.TaskInfo(
        data=dumped,
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))
    task = decode(proto)

    assert isinstance(task, PythonTask)
    assert task['data'] == dumped
    assert task.data == data

    proto = mesos_pb2.TaskInfo(
        labels=mesos_pb2.Labels(
            labels=[mesos_pb2.Label(key='python')]))
    task = decode(proto)
    task.data = data

    assert isinstance(task, PythonTask)
    assert task.data == data
    assert task['data'] == dumped


def test_python_task_encode():
    fn, args, kwargs = sum, [range(5)], {}
    data = (fn, args, kwargs)
    dumped = cloudpickle.dumps(data)

    task = PythonTask(fn=fn, args=args, kwargs=kwargs,
                      id={'value': 'test-id'})

    proto = encode(task)
    assert isinstance(proto, mesos_pb2.TaskInfo)
    assert proto.data == dumped
    assert proto.task_id.value == 'test-id'

    task = PythonTask(id=TaskID(value='test-id'))
    task.data = data
    proto = encode(task)
    assert isinstance(proto, mesos_pb2.TaskInfo)
    assert proto.data == dumped
    assert proto.task_id.value == 'test-id'


def test_python_task_execution():
    fn, args, kwargs = sum, [range(5)], {}
    task = PythonTask(fn=fn, args=args, kwargs=kwargs,
                      id={'value': 'test-id'})
    task = decode(encode(task))
    assert task() == 10

    def fn(lst1, lst2):
        return sum(lst1) - sum(lst2)
    args = [range(5), range(3)]
    task = PythonTask(fn=fn, args=args, id={'value': 'test-id'})
    task = decode(encode(task))
    assert task() == 7


# def test_python_task_callbacks(mocker):
#     fn, args, kwargs = sum, [range(5)], {}

#     on_update = mocker.MagicMock()
#     on_success = mocker.MagicMock()
#     task = PythonTask(fn=fn, args=args, kwargs=kwargs, id={'value': 'test-id'},
#                       on_update=on_update, on_success=on_success)
#     status = PythonTaskStatus(state='TASK_FINISHED', data=20)
#     task.update(status)
#     on_update.assert_called_with(task, status)
#     on_success.assert_called_with(task, status)

#     on_update = mocker.MagicMock()
#     on_fail = mocker.MagicMock()
#     task = PythonTask(fn=fn, args=args, kwargs=kwargs, id={'value': 'test-id'},
#                       on_update=on_update, on_fail=on_fail)
#     status = PythonTaskStatus(state='TASK_KILLED')
#     task.update(status)
#     on_update.assert_called_with(task, status)
#     on_fail.assert_called_with(task, status)



from __future__ import absolute_import, division, print_function

from .scheduler import SchedulerProxy, SchedulerDriverProxy
from .executor import ExecutorProxy, ExecutorDriverProxy
from .messages import MessageProxy


__all__ = ('MessageProxy',
           'SchedulerProxy',
           'ExecutorProxy',
           'SchedulerDriverProxy',
           'ExecutorDriverProxy')
